/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "gnu-pop3d.h"

#define LINESLICE 256

/* Prints out the specified message */
int
pop3_retr (const char *arg)
{
  int mesg, empty_line = 1;
  int done = 0;
  int lbuf;
  char buf[LINESLICE * sizeof (char)], lineend[3 * sizeof (char)];
  int header = 1;
  char *returnpath = NULL, *envto = NULL, *delivto = NULL, *sender =
	NULL, *msgid = NULL, *date = NULL, *subject = NULL;
  char *res;

  if ((strlen (arg) == 0) || (strchr (arg, ' ') != NULL))
	return ERR_BAD_ARGS;
  if (state != TRANSACTION)
	return ERR_WRONG_STATE;
  mesg = atoi (arg) - 1;
  if (pop3_mesg_exist (mesg) != OK)
	return ERR_NO_MESG;
  if (fsetpos (mbox, &(messages[mesg].header)) < 0)
	return ERR_FILEPOS;
  fprintf (ofile, "+OK %u octets\r\n", messages[mesg].size);
  while ((res = fgets (buf, LINESLICE, mbox)) && !done)
	{
	  if (empty_line && !strncmp (buf, "From ", 5))
		done = 1;

	  else
		{
		  lineend[0] = '\0';
		  if ((buf[0] == '\n') || (buf[0] == '\r') || (buf[0] == '\0'))
			empty_line = 1;

		  else
			empty_line = 0;
		  if (empty_line == 1)
			header = 0;
		  if ((mdn != 0) && header)
			{
			  if (!returnpath)
				returnpath = pop3_pickupheader (buf, "Return-Path");
			  if (!subject)
				subject = pop3_pickupheader (buf, "Subject");
			  if (!date)
				date = pop3_pickupheader (buf, "Date");
			  if (!msgid)
				msgid = pop3_pickupheader (buf, "Message-ID");
			  if (!envto)
				envto = pop3_pickupheader (buf, "X-Original-To");
			  if (!envto)
				envto = pop3_pickupheader (buf, "Envelope-To");
			  if (!envto)
				envto = pop3_pickupheader (buf, "X-Envelope-To");
			  if (!delivto && !envto)
				delivto = pop3_pickupheader (buf, "Delivered-To");
			  if (!sender)
				sender = pop3_pickupheader (buf, "Disposition-Notification-To");
			  if (!sender)
				sender = pop3_pickupheader (buf, "Return-Receipt-To");
			}
		  lbuf = strlen (buf);
		  if (buf[lbuf - 1] == '\n')
			{
			  lbuf--;
			  strcpy (lineend, "\r\n");
			}
		  if (buf[lbuf - 1] == '\r')
			lbuf--;

/*		if ((buf[0] == '.') && ((lbuf == 1) || (buf[1] == ' '))) */
		  if (buf[0] == '.')
			fprintf (ofile, ".%.*s%s", lbuf, buf, lineend);

		  else
			fprintf (ofile, "%.*s%s", lbuf, buf, lineend);
		  while ((res != NULL) && (lineend[0] == '\0' && strchr (buf, '\n') == NULL))
			{
			  res = fgets (buf, LINESLICE, mbox);
			  if (res == NULL)
				{
				  buf[0] = '\n';
				  buf[1] = '\0';
				}				/* last line in mailbox without \n */
			  lbuf = strlen (buf);
			  if (buf[lbuf - 1] == '\n')
				{
				  lbuf--;
				  strcpy (lineend, "\r\n");
				}
			  if ((lbuf > 0) && (buf[lbuf - 1] == '\r'))
				lbuf--;
			  fprintf (ofile, "%.*s%s", lbuf, buf, lineend);
			}
		}
	}
  fprintf (ofile, ".\r\n");
  if (mdn)
	{
	  if (sender && !strchr (messages[mesg].status, 'R'))
		{
		  pop3_sendmdn (returnpath, envto, delivto, subject, date, msgid, sender);
		  mdncount++;
		}
	  if (returnpath)
		free (returnpath);
	  if (subject)
		free (subject);
	  if (date)
		free (date);
	  if (msgid)
		free (msgid);
	  if (envto)
		free (envto);
	  if (delivto)
		free (delivto);
	  if (sender)
		free (sender);
	}
  if (strchr (messages[mesg].status, 'R') == NULL)
	{
	  messages[mesg].chng = 1;
	  strcat (messages[mesg].status, "R");
	}
  return OK;
}
